#ifdef HAVE_CONFIG_H
# include "config.h"     
#endif
#include <iostream>
#include <string>
#include <fstream>
#include <dune/common/parallel/mpihelper.hh> // An initializer of MPI
#include <dune/common/exceptions.hh> // We use exceptions
#include <dune/grid/alugrid.hh> // We use the ALUGrid manager
#include <dune/grid/alugrid/3d/alugrid.hh> // We use the 3D ALUGrid manager
#include <dune/grid/alugrid/common/backuprestore.hh>
#include <dune/grid/io/file/vtk/vtkwriter.hh>


int main(int argc, char** argv)
{
  try{
    //Maybe initialize Mpi
    Dune::MPIHelper& helper = Dune::MPIHelper::instance(argc, argv);
    std::cout << "Hello World! This is adaptFromRestore." << std::endl;
    if(Dune::MPIHelper::isFake)
      std::cout<< "This is a sequential program." << std::endl;
    else
      std::cout<<"I am rank "<<helper.rank()<<" of "<<helper.size()
        <<" processes!"<<std::endl;

    typedef Dune::ALUGrid<3,3,Dune::cube,Dune::nonconforming> GridType;
    typedef GridType::ctype ct;
    const int dim = GridType::dimension;
    typedef Dune::FieldVector< ct,dim > DomainType;
    typedef GridType::LeafGridView LeafGridView;
    typedef LeafGridView::Codim<0>::Partition<Dune::All_Partition>::Iterator LeafElementIterator;
    typedef LeafElementIterator::Entity LeafElement;
    typedef Dune::LeafMultipleCodimMultipleGeomTypeMapper< GridType,Dune::MCMGElementLayout > LeafElementMapperType;

    GridType restoreGrid;
    
    //RESTORE GRID
    std::stringstream outputPathStream;
    std::string outputPath = "./data_";
    outputPathStream << outputPath << helper.size() << "/";
    outputPath = outputPathStream.str();
    std::string filename = "writtenGrid.";
    double tt = 0.0;
    std::stringstream filenameStream;
    filenameStream << outputPath << filename << helper.rank();
    restoreGrid.readGrid<Dune::xdr>(filenameStream.str(),tt);

    //CONTROL GRID BY VTK PLOT
    LeafGridView leafGridView = restoreGrid.leafView();
    Dune::VTKWriter<LeafGridView> vtkwriter(leafGridView);
    filename = "restoredGrid";
    vtkwriter.pwrite(filename,outputPath,"",Dune::VTK::ascii);

    //MAPPER AND DATA DECLARATION
    Dune::MCMGElementLayout<dim> layout;
    LeafElementMapperType leafElementMapper(restoreGrid,layout);
    int mapperSize = leafElementMapper.size();
    std::vector<std::vector<int> > marks(mapperSize,std::vector<int>(3,0));
    std::vector<std::string> names(3);
    names[0] = "mark"; names[1] = "index"; names[2] = "rank";
    
    //OPEN INPUT FILE STREAM
    filenameStream.str("");
    filename = "Marks.";
    filenameStream << outputPath << filename << helper.rank();
    std::ifstream in(filenameStream.str().c_str());

    //RESTORE DATA
    int index;
    while(in.good())
      {
	in >> index;
	for(int i = 0; i < 3; ++i)
	  {
	    in >> marks[index][i];
	  }
      }
    in.close();

    //CONTROL DATA BY VTK PLOT
    std::vector<std::vector<int> > tempField(3,std::vector<int>(mapperSize,0));
    for(int i = 0; i < 3; i++)
      {
	filenameStream.str("");
	filenameStream << names[i];
	for(int j = 0; j < mapperSize; ++j)
	  tempField[i][j] = marks[j][i];
	vtkwriter.addCellData(tempField[i],filenameStream.str().c_str());
      }
    filenameStream.str("");
    filenameStream << "Data";
    vtkwriter.pwrite(filenameStream.str(),outputPath,"",Dune::VTK::ascii);
    

    //MARK ELEMENT
    LeafElementIterator end = leafGridView.end<0,Dune::All_Partition>();
    for(LeafElementIterator it = leafGridView.begin<0,Dune::All_Partition>(); it != end; ++it)
      {
	const LeafElement & element = *it;
	int indexi = leafElementMapper.map(element);
	restoreGrid.mark(marks[indexi][0],element);
      }
    
    //ADAPT GRID
    restoreGrid.preAdapt();
    restoreGrid.adapt();
    restoreGrid.postAdapt();

    //CONTROL ADAPTATION BY VTK PLOT
    leafGridView = restoreGrid.leafView();
    Dune::VTKWriter<LeafGridView> vtkwriter2(leafGridView);
    filename = "adaptedGrid";
    vtkwriter2.pwrite(filename,outputPath,"",Dune::VTK::ascii);


    return 0;
  }
  catch (Dune::Exception &e){
    std::cerr << "Dune reported error: " << e << std::endl;
  }
  catch (...){
    std::cerr << "Unknown exception thrown!" << std::endl;
  }
}
